// TEView.cpp : implementation of the CTEditView class
//

#include "stdafx.h"
#include <Winuser.h>
#include "TEdit.h"
#include "TEDoc.h"
#include "TEView.h"
#include "MainFrm.h"
#include "SaveQDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

static const int TAB_DX = 4;

// Description of editor commands

// Bits in state field of command descriptor
const UINT SHIFTL_MASK = 3;     // Bits 0, 1
const UINT CONTROL_MASK = 12;   // Bits 2, 3

const int MODIFIERS_ANY = 0;

const UINT SHIFT_ANY = 0;
const UINT SHIFT_OFF = 1;
const UINT SHIFT_ON = 2;

const UINT CONTROL_ANY = 0;
const UINT CONTROL_OFF = 4;
const UINT CONTROL_ON = 8;

// Table of editors commands
const struct CTEditView::CommandDsc CTEditView::editorCommands[] =
{
    {VK_DOWN, CONTROL_OFF, false, &CTEditView::onDown},   // Down arrow
    {VK_DOWN, CONTROL_ON, false, &CTEditView::onTextEnd}, // Ctrl+Down
    {VK_NUMPAD2, CONTROL_OFF, false, &CTEditView::onDown},// Down arrow on keypad
    {VK_NUMPAD2, CONTROL_ON, false, &CTEditView::onTextEnd},// Ctrl+Down on keypad
    {VK_UP, CONTROL_OFF, false, &CTEditView::onUp},         // Up arrow
    {VK_UP, CONTROL_ON, false, &CTEditView::onTextBeg},     // Ctrl+Up
    {VK_NUMPAD8, CONTROL_OFF, false, &CTEditView::onUp},    // Up arrow on keypad
    {VK_NUMPAD8, CONTROL_ON, false, &CTEditView::onTextBeg},// Ctrl+Up on keypad
    {VK_LEFT, CONTROL_OFF, false, &CTEditView::onLeft},     // Left arrow
    {VK_LEFT, CONTROL_ON, false, &CTEditView::onHome},      // Ctrl+Left
    {VK_NUMPAD4, CONTROL_OFF, false, &CTEditView::onLeft},  // Left arrow on keypad
    {VK_NUMPAD4, CONTROL_ON, false, &CTEditView::onHome},   // Ctrl+Left on keypad
    {VK_RIGHT, CONTROL_OFF, false, &CTEditView::onRight},   // Right arrow
    {VK_RIGHT, CONTROL_ON, false, &CTEditView::onEnd},      // Ctrl+Right
    {VK_NUMPAD6, CONTROL_OFF, false, &CTEditView::onRight}, // Right arrow on keypad
    {VK_NUMPAD6, CONTROL_ON, false, &CTEditView::onEnd},    // Ctrl+Right on keypad
    {VK_HOME, CONTROL_OFF, false, &CTEditView::onHome},     // Home
    {VK_HOME, CONTROL_ON, false, &CTEditView::onTextBeg},   // Ctrl+Home
    {VK_NUMPAD7, CONTROL_OFF, false, &CTEditView::onHome},  // Home on keypad
    {VK_NUMPAD7, CONTROL_ON, false, &CTEditView::onTextBeg},// Ctrl+Home on keypad
    {VK_END, CONTROL_OFF, false, &CTEditView::onEnd},       // End
    {VK_END, CONTROL_ON, false, &CTEditView::onTextEnd},    // Ctrl+End
    {VK_NUMPAD1, CONTROL_OFF, false, &CTEditView::onEnd},   // End on keypad
    {VK_NUMPAD1, CONTROL_ON, false, &CTEditView::onTextEnd},
    {VK_RETURN, MODIFIERS_ANY, true, &CTEditView::onEnter}, // Enter
    {VK_DELETE, SHIFT_OFF, true, &CTEditView::onDelete},    // Delete
    {VK_DELETE, SHIFT_ON, true, &CTEditView::onDeleteLine}, // Shift+Delete
    {VK_INSERT, SHIFT_OFF, true, &CTEditView::onInsert},    // Insert
    {VK_INSERT, SHIFT_ON, true, &CTEditView::onInsertLine}, // Shift+Insert
    {VK_BACK, MODIFIERS_ANY, true, &CTEditView::onBackSpace}, // Back Space
    {VK_PRIOR, CONTROL_OFF, false, &CTEditView::onPageUp},    // PgUp
    {VK_PRIOR, CONTROL_ON, false, &CTEditView::onTextBeg},    // Ctrl+PgUp
    {VK_NUMPAD9, CONTROL_OFF, false, &CTEditView::onPageUp},  // PgUp on keypad
    {VK_NUMPAD9, CONTROL_ON, false, &CTEditView::onTextBeg},  // Ctrl+PgUp on keypad
    {VK_NEXT, CONTROL_OFF, false, &CTEditView::onPageDown},   // PgDn
    {VK_NEXT, CONTROL_ON, false, &CTEditView::onTextEnd},     // Ctrl+PgDn
    {VK_NUMPAD3, CONTROL_OFF, false, &CTEditView::onPageDown},// PgDn on keypad
    {VK_NUMPAD9, CONTROL_ON, false, &CTEditView::onTextEnd},  // Ctrl+PgDn on keypad
    {VK_TAB, CONTROL_OFF|SHIFT_OFF, false, &CTEditView::onTab},   // Tab
    {VK_TAB, CONTROL_OFF|SHIFT_ON, false, &CTEditView::onTabLeft},// Shift+Tab
    {VK_TAB, CONTROL_ON, true, &CTEditView::onInsertTab},     // Ctrl+Tab

    {'S', CONTROL_ON, false, &CTEditView::OnFileSave},  // Ctrl+S
    {'Q', CONTROL_ON, false, &CTEditView::onQuit},      // Ctrl+Q
    {'N', CONTROL_ON, false, &CTEditView::OnFileNew},   // Ctrl+N
    {'N', CONTROL_ON, false, &CTEditView::OnFileOpen},  // Ctrl+O
    {'K', CONTROL_ON, true, &CTEditView::onDeleteLine}, // Ctrl+K
    {'L', CONTROL_ON, true, &CTEditView::onInsertLine}, // Ctrl+L

    {0, 0, false, 0}    // Terminator
};

// CTEditView

IMPLEMENT_DYNCREATE(CTEditView, CView)

BEGIN_MESSAGE_MAP(CTEditView, CView)
    // Standard printing commands
    ON_COMMAND(ID_FILE_PRINT, &CView::OnFilePrint)
    ON_COMMAND(ID_FILE_PRINT_DIRECT, &CView::OnFilePrint)
    ON_COMMAND(ID_FILE_PRINT_PREVIEW, &CView::OnFilePrintPreview)
    ON_WM_CHAR()
    ON_COMMAND(ID_FILE_OPEN, &CTEditView::OnFileOpen)
    ON_COMMAND(ID_FILE_NEW, &CTEditView::OnFileNew)
    ON_COMMAND(ID_FILE_SAVE, &CTEditView::OnFileSave)
    ON_UPDATE_COMMAND_UI(ID_FILE_SAVE, &CTEditView::OnUpdateFileSave)
    ON_COMMAND(ID_FILE_SAVE_AS, &CTEditView::OnFileSaveAs)
    ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_AS, &CTEditView::OnUpdateFileSaveAs)
    ON_WM_SIZE()
    ON_WM_ERASEBKGND()
    ON_WM_LBUTTONDOWN()
END_MESSAGE_MAP()

// CTEditView construction/destruction

CTEditView::CTEditView():
    lastCharTyped(0),
    cursorX(0),
    cursorY(0),
    inputDisabled(false),
    windowX(0),
    windowY(0),
    windowWidth(80),
    windowHeight(25),
    leftMargin(4),
    topMargin(4),
    rightMargin(4),
    bottomMargin(4),

    fontParametersSet(false),
    charWidth(16),
    charHeight(24),
    leading(8),
    dx(charWidth),
    dy(charHeight + leading),
    endOfText("[* End of text *]"),
    textFont(),
    fontLoaded(false),
    bgBrush(RGB(235, 235, 230)),
    cursorBgBrush(RGB(80, 0, 160)),
    message("Ready"),
    m_bShiftOn(false),
    m_bControlOn(false)
{
    // TODO: add construction code here
    ((CTEditApp*) AfxGetApp())->m_pView = this;
}

CTEditView::~CTEditView()
{
    ((CTEditApp*) AfxGetApp())->m_pView = NULL;
}

BOOL CTEditView::PreCreateWindow(CREATESTRUCT& cs)
{
    // TODO: Modify the Window class or styles here by modifying
    //  the CREATESTRUCT cs

    return CView::PreCreateWindow(cs);
}

// CTEditView drawing

void CTEditView::OnDraw(CDC* pDC)
{
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return;

    wchar_t outputLine[512]; // Buffer for UNICODE line for text output

    CRect clientRect;
    GetClientRect(clientRect);
    pDC->FillRect(clientRect, &bgBrush);

    CFont* oldFont = pDC->SelectObject(&textFont);

    CSize s;
    if (!fontParametersSet && fontLoaded)
    {
        s = pDC->GetOutputTextExtent(_T("W"));
        charWidth = pDC->GetOutputTextExtent(_T("WW")).cx - s.cx;
        s = pDC->GetOutputTextExtent(_T("Ay"));
        charHeight = s.cy;
        leading = charHeight / 8;
        if (leading < 4)
            leading = 4;
        if (leading > 32)
            leading = 32;
        dx = charWidth;
        if (dx <= 0)
            dx = 4;
        dy = charHeight + leading;
        //... windowWidth = (clientRect.Width() - leftMargin - rightMargin)
        windowWidth = (clientRect.Width() - leftMargin)
            / dx;
        //... windowHeight =
        //...     (clientRect.Height() + leading - topMargin - bottomMargin)
        windowHeight =
            (clientRect.Height() + leading - topMargin)
            / dy;
        fontParametersSet = true;
    }

    // TODO: add draw code for native data here
    pDC->SetBkMode(TRANSPARENT);
    pDC->SetTextColor(RGB(0, 0, 0));    // Black

    // Draw a text in a window
    int x = leftMargin;
    int y = topMargin;
    for (int screenY = 0; screenY < windowHeight; ++screenY, y += dy) {
        int textY = windowY + screenY;
        const TextLine* currentLine;
        if (textY > pDoc->text.size()) {
            break;
        } else if (textY == pDoc->text.size()) {
            currentLine = &(endOfText);
        } else {
            currentLine = &(pDoc->text.getLine(textY));
        }
        int len = currentLine->length();
        if (len > windowX)
        {
            int restrictedLen = len - windowX;
            if (restrictedLen > windowWidth)
                restrictedLen = windowWidth;
            if (restrictedLen > 0)
            {
                const char *visibleLine =
                    currentLine->getString() + windowX;

                /*...
                s = pDC->GetOutputTextExtent(
                    visibleLine, restrictedLen
                );
                pDC->DrawText(
                    visibleLine,
                    restrictedLen,
                    CRect(x, y, x + s.cx, y + s.cy),
                    DT_LEFT | DT_TOP        //DT_VCENTER
                );
                ...*/

                if (restrictedLen > 512)
                    restrictedLen = 512;

                // Convert to UNICODE line
                int res = MultiByteToWideChar(
                    pDoc->codepage,
                    MB_PRECOMPOSED,
                    visibleLine,
                    restrictedLen,
                    outputLine,
                    512
                );

                if (res > 0)
                {
                    s = pDC->GetOutputTextExtent(
                        outputLine, res
                    );
                    pDC->DrawText(
                        outputLine,
                        res,
                        CRect(
                            x, y,
                            x + s.cx,
                            y + s.cy
                        ),
                        DT_LEFT | DT_TOP        //DT_VCENTER
                    );
                }
                else
                {
                    TRACE("Error convertion to UNICODE.\n");
                }
            }
        }
    }

    // Draw cursor
    if (!inputDisabled)
        DrawCursor(cursorX, cursorY, true, pDC);

    pDC->SelectObject(oldFont);
}


// CTEditView printing

BOOL CTEditView::OnPreparePrinting(CPrintInfo* pInfo)
{
    // default preparation
    return DoPreparePrinting(pInfo);
}

void CTEditView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
    // TODO: add extra initialization before printing
}

void CTEditView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
    // TODO: add cleanup after printing
}


// CTEditView diagnostics

#ifdef _DEBUG
void CTEditView::AssertValid() const
{
    CView::AssertValid();
}

void CTEditView::Dump(CDumpContext& dc) const
{
    CView::Dump(dc);
}

CTEditDoc* CTEditView::GetDocument() const // non-debug version is inline
{
    ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CTEditDoc)));
    return (CTEditDoc*)m_pDocument;
}
#endif //_DEBUG


// CTEditView message handlers

void CTEditView::OnInitialUpdate()
{
    CView::OnInitialUpdate();

    // TODO: Add your specialized code here and/or call the base class
    LoadFont();
}

void CTEditView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
    // TODO: Add your message handler code here and/or call default
    TRACE("OnChar: nChar=0x%x, nFlags=0x%x\n", nChar, nFlags);
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return;

    // State of shift and control keys
    m_bShiftOn = (
        (0x8000 & GetKeyState(VK_SHIFT)) != 0
    );
    m_bControlOn = (
        (0x8000 & GetKeyState(VK_CONTROL)) != 0
    );

    if (m_bControlOn)   // Control chars should be processed as commands
        return;

    wchar_t line[2];
    line[0] = nChar; line[1] = 0;
    char outputLine[8];
    BOOL useDefault = TRUE;

    int res = WideCharToMultiByte(
        pDoc->codepage,
        0,                   // dwFlags - performance and mapping flags
        line,               // wide-character string
        1,                   // number of chars in string.
        outputLine,         // buffer for new string
        8,                  // size of buffer
        " ",                // default for unmappable chars
        &useDefault         // set when default char used
    );
    if (res <= 0)
        return;

    lastCharTyped =  outputLine[0];
    lastCharTyped &= 0xff;  // Clear the high byte

    PreProcessCommand();
    onCharTyped();
    pDoc->textChanged = true;
    PostProcessCommand();

    //... CView::OnChar(nChar, nRepCnt, nFlags);
}

void CTEditView::OnFileOpen()
{
    // TODO: Add your command handler code here
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return;

    if (pDoc->textChanged)
    {
        CSaveQueryDlg saveDlg;
        UINT res = saveDlg.DoModal();
        if (res == IDOK)
        {
            OnFileSaveAs();
        }
        else if (res == IDCANCEL)
            return;
    }

    CFileDialog openDlg(TRUE);

    const TCHAR filter[] = _T("Text files\0*.*\0\0\0");
    openDlg.m_ofn.lpstrFilter = filter;

    CString initialDir, initialFile;
    TCHAR initDir[1024];
    TCHAR initFile[1024];
    initDir[0] = 0;
    initFile[0] = 0;
    if (pDoc->filePathSet)
        _tcsncpy(initDir, pDoc->fileDirectory, 1022);

    openDlg.m_ofn.lpstrTitle = _T("Open a file with plain text");

    if (openDlg.DoModal() != IDOK)
        return;

    pDoc->text.load(openDlg.GetPathName());
    //... pDoc->filePath = openDlg.GetPathName();
    //... pDoc->filePathSet = true;
    pDoc->SetFilePath(openDlg.GetPathName());
    pDoc->SetTitle(pDoc->filePath);
    pDoc->textChanged = false;
    pDoc->textSaved = false;
    AfxGetApp()->AddToRecentFileList(pDoc->filePath);

    CMainFrame* pFrame = (CMainFrame*) GetParent();
    pFrame->SetWindowText(pDoc->filePath);
    pFrame->SetMessageText(_T("Edit a text"));

    cursorX = 0;
    cursorY = 0;
    windowX = 0;
    windowY = 0;

    Invalidate();
}

void CTEditView::OnFileNew()
{
    // TODO: Add your command handler code here
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return;

    if (pDoc->textChanged)
    {
        CSaveQueryDlg saveDlg;
        UINT res = saveDlg.DoModal();
        if (res == IDOK)
        {
            OnFileSave();
        }
        else if (res == IDCANCEL)
            return;
    }

    cursorX = 0;
    cursorY = 0;
    windowX = 0;
    windowY = 0;

    pDoc->text.clear();
    pDoc->textChanged = FALSE;
    pDoc->textSaved = FALSE;
    pDoc->filePathSet = FALSE;
    pDoc->SetTitle(_T("NoName"));

    Invalidate();
}

void CTEditView::OnFileSave()
{
    // TODO: Add your command handler code here
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return;

    if (!(pDoc->filePathSet))
    {
        OnFileSaveAs();
        return;
    }

    if (!(pDoc->text.save(pDoc->filePath)))
    {
        AfxMessageBox(_T("Could not save a text."));
    }
    else
    {
        pDoc->textChanged = false;
        pDoc->textSaved = true;
    }
}

void CTEditView::OnUpdateFileSave(CCmdUI *pCmdUI)
{
    // TODO: Add your command update UI handler code here
    CTEditDoc* pDoc = GetDocument();
    pCmdUI->Enable(
        pDoc != NULL &&
        pDoc->textChanged
    );
}

void CTEditView::OnFileSaveAs()
{
    // TODO: Add your command handler code here
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return;

    CFileDialog openDlg(FALSE);

    const TCHAR filter[] = _T("Text files\0*.*\0\0\0");
    openDlg.m_ofn.lpstrFilter = filter;
    openDlg.m_ofn.nFilterIndex = 0;
    openDlg.m_ofn.lpstrDefExt = _T("");

    CString initialDir, initialFile;
    TCHAR initDir[1024];
    TCHAR initFile[1024];
    initDir[0] = 0;
    initFile[0] = 0;
    if (pDoc->filePathSet)
    {
        _tcsncpy(initDir, pDoc->fileDirectory, 1022);
        _tcsncpy(initFile, pDoc->fileName, 1022);
    }

    openDlg.m_ofn.lpstrFile = initFile;
    openDlg.m_ofn.nMaxFile = 1024;
    openDlg.m_ofn.lpstrInitialDir = initDir;
    openDlg.m_ofn.lpstrTitle = _T("Save a text");

    if (openDlg.DoModal() != IDOK)
        return;

    if (!pDoc->text.save(openDlg.GetPathName()))
    {
        AfxMessageBox(_T("Write error"));
    }
    else
    {
        //... pDoc->filePath = openDlg.GetPathName();
        //... pDoc->filePathSet = true;
        pDoc->SetFilePath(openDlg.GetPathName());
        pDoc->textChanged = false;
        pDoc->textSaved = true;
        pDoc->SetTitle(pDoc->filePath);
    }
}

void CTEditView::OnUpdateFileSaveAs(CCmdUI *pCmdUI)
{
    // TODO: Add your command update UI handler code here
}

void CTEditView::OnSize(UINT nType, int cx, int cy)
{
    CView::OnSize(nType, cx, cy);

    // TODO: Add your message handler code here
    if (fontParametersSet)
    {
        CRect clientRect;
        GetClientRect(clientRect);
        windowWidth = (clientRect.Width() - leftMargin)
            / dx;
        windowHeight = (clientRect.Height() + leading - topMargin)
            / dy;
    }
}

void CTEditView::LoadFont()
{
    if (!fontLoaded)
    {
        textFont.CreateFont(
           20,                        // nHeight
           0,                         // nWidth
           0,                         // nEscapement
           0,                         // nOrientation
           FW_NORMAL,                 // nWeight
           FALSE,                     // bItalic
           FALSE,                     // bUnderline
           0,                         // cStrikeOut
           ANSI_CHARSET,              // nCharSet
           OUT_DEFAULT_PRECIS,        // nOutPrecision
           CLIP_DEFAULT_PRECIS,       // nClipPrecision
           DEFAULT_QUALITY,           // nQuality
           DEFAULT_PITCH | FF_SWISS,  // nPitchAndFamily
           _T("Courier New")          // lpszFacename
        );
        fontLoaded = true;
    }
}

BOOL CTEditView::OnEraseBkgnd(CDC* pDC)
{
    // TODO: Add your message handler code here and/or call default
    CRect clientRect;
    GetClientRect(clientRect);
    pDC->FillRect(clientRect, &bgBrush);
    return TRUE;
}

bool CTEditView::ProcessKeydownMessage(UINT nChar, UINT nFlags)
{
    CTEditDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
    if (pDoc == NULL)
        return false;

    // State of shift and control keys
    m_bShiftOn = (
        (0x8000 & GetKeyState(VK_SHIFT)) != 0
    );
    m_bControlOn = (
        (0x8000 & GetKeyState(VK_CONTROL)) != 0
    );


    // Look up the command in the table
    const struct CommandDsc* command = editorCommands;
    bool commandFound = false;
    while (!commandFound && command->method != 0)
    {
        UINT shiftState = (SHIFTL_MASK & command->state);
        UINT controlState = (CONTROL_MASK & command->state);
        if (
            nChar == command->virtualKey
            &&
            (
                shiftState == SHIFT_ANY ||
                (!m_bShiftOn && shiftState == SHIFT_OFF) ||
                (m_bShiftOn && shiftState == SHIFT_ON)
            )
            &&
            (
                controlState == CONTROL_ANY ||
                (!m_bControlOn && controlState == CONTROL_OFF) ||
                (m_bControlOn && controlState == CONTROL_ON)
            )
        )
        {
            commandFound = true; break;
        }
        ++command;
    }

    if (commandFound)
    {
        PreProcessCommand();
        (this->*(command->method))();   // Perform the command
        if (command->change) {
            pDoc->textChanged = true;   // Command changes the text
        }
        PostProcessCommand();
        return true;
    }
    else
    {
        return false;
    }
}

void CTEditView::onDown()   // Down arrow
{
    CTEditDoc* pDoc = GetDocument();
    if (cursorY < pDoc->text.size())
        ++cursorY;
}

void CTEditView::onTextEnd() // Ctrl+Down
{
    CTEditDoc* pDoc = GetDocument();
    cursorX = 0;
    cursorY = pDoc->text.size();
}

void CTEditView::onUp() // Up arrow
{
    if (cursorY > 0)
        --cursorY;
}

void CTEditView::onTextBeg() // Ctrl+Up
{
    cursorX = 0;
    cursorY = 0;
}

void CTEditView::onLeft() {
    if (cursorX > 0)
        --cursorX;
}

void CTEditView::onRight() {
    ++cursorX;
}

void CTEditView::onHome() {
    cursorX = 0;
}

void CTEditView::onEnd() {
    CTEditDoc* pDoc = GetDocument();
    if (cursorY >= pDoc->text.size()) {
        cursorX = 0;
    } else {
        cursorX = pDoc->text.getLine(cursorY).length();
    }
}

void CTEditView::DrawCursor(
    int cx, int cy,
    bool cursorOn /* = true */
)
{
    CClientDC dc(this);
    DrawCursor(cx, cy, cursorOn, &dc);
}

void CTEditView::DrawCursor(
    int cx, int cy, bool cursorOn, CDC* pDC
)
{
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return;

    if (
        cx < windowX || cy < windowY ||
        cx >= windowX + windowWidth ||
        cy >= windowY + windowHeight
    )
        return;         // Cursor outside of window

    ASSERT(pDC != NULL);

    int x = leftMargin + (cx - windowX) * dx;
    int y = topMargin + (cy - windowY) * dy;

    CBrush* bg = &cursorBgBrush;
    if (!cursorOn)
        bg = &bgBrush;

    pDC->FillRect(
        CRect(x, y, x+charWidth, y+charHeight),
        bg
    );

    if (cy <= pDoc->text.size())
    {
        TextLine* line;
        if (cy == pDoc->text.size()) {
            line = &(endOfText);
        } else {
            line = &(pDoc->text.getLine(cy));
        }

        if (cx < line->length())
        {
            CFont* oldFont = pDC->SelectObject(&textFont);

            pDC->SetBkMode(TRANSPARENT);
            if (cursorOn)
                pDC->SetTextColor(RGB(255, 255, 255));  // White
            else
                pDC->SetTextColor(RGB(0, 0, 0));    // Black

            // Convert to UNICODE line
            wchar_t outputLine[2];  // Buffer for UNICODE line for text output

            int res = MultiByteToWideChar(
                pDoc->codepage,
                MB_PRECOMPOSED,
                line->getString() + cx,
                1,
                outputLine,
                2
            );
            outputLine[1] = 0;
            if (res > 0)
            {
                pDC->DrawText(
                    outputLine,
                    1,
                    CRect(
                        x, y,
                        x + charWidth,
                        y + charHeight
                    ),
                    DT_LEFT | DT_TOP        //DT_VCENTER
                );
            }
            pDC->SelectObject(oldFont);
        }
    }
}

void CTEditView::PreProcessCommand()
{
    inputDisabled = true; // Disable any input while command is not completed
    DrawCursor(cursorX, cursorY, false); // Remove cursor
}

void CTEditView::PostProcessCommand()
{
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return;

    // Scroll the window to the cursor, if necessary
    if (
        cursorX < windowX || cursorX > windowX + windowWidth - 1 ||
        cursorY < windowY || cursorY > windowY + windowHeight - 1
    ) {
        // Cursor is removed at the moment!
        ScrollToCursor();
    }

    // Set pointer in the text
    pDoc->text.setPointer(cursorY);

    inputDisabled = false;

    DrawCursor(cursorX, cursorY, true);
}

void CTEditView::ScrollToCursor()
{
    if (cursorX < windowX) {
        ScrollLeft(windowX - cursorX);
    } else if (cursorX > windowX + windowWidth - 1) {
        ScrollRight(cursorX - (windowX + windowWidth - 1));
    }

    if (cursorY < windowY) {
        ScrollUp(windowY - cursorY);
    } else if (cursorY > windowY + windowHeight - 1) {
        ScrollDown(cursorY - (windowY + windowHeight - 1));
    }
}

void CTEditView::ScrollLeft(int n) {
    if (n == 0)
        return;
    if (n > windowX)
        n = windowX;
    windowX -= n;
    if (n > windowWidth / 2) {
        Invalidate();
    } else {
        int shift = dx * n;
        CRect clientRect;
        GetClientRect(clientRect);
        CRect textRect( // Clippng rectangle
            leftMargin, topMargin,
            leftMargin + windowWidth*dx,
            topMargin + windowHeight*dy
        );
        ScrollWindow(
            shift, 0,
            textRect,   // Rectangle to scroll
            textRect    // Clippng rectangle
        );
        InvalidateRect(
            CRect(
                0, 0,
                leftMargin + shift,
                clientRect.bottom
            )
        );
    }
}

void CTEditView::ScrollRight(int n) {
    if (n == 0)
        return;
    windowX += n;
    if (n > windowWidth / 2) {
        Invalidate();
    } else {
        int shift = dx * n;
        CRect clientRect;
        GetClientRect(clientRect);
        CRect textRect( // Clippng rectangle
            leftMargin, topMargin,
            leftMargin + windowWidth*dx,
            topMargin + windowHeight*dy
        );
        ScrollWindow(
            -shift, 0,
            textRect,   // Rectangle to scroll
            textRect    // Clippng rectangle
        );
        InvalidateRect(
            CRect(
                leftMargin + windowHeight * dx - shift,
                0,
                clientRect.right,
                clientRect.bottom
            )
        );
    }
}

void CTEditView::ScrollUp(int n) {
    if (n == 0)
        return;
    if (n > windowY)
        n = windowY;
    windowY -= n;

    if (n > windowWidth / 2) {
        Invalidate();
    } else {
        int shift = dy * n;
        CRect clientRect;
        GetClientRect(clientRect);
        CRect textRect( // Clippng rectangle
            leftMargin, topMargin,
            leftMargin + windowWidth*dx,
            topMargin + windowHeight*dy
        );
        ScrollWindow(
            0, shift,
            textRect,   // Rectangle to scroll
            textRect    // Clippng rectangle
        );
        InvalidateRect(
            CRect(
                0, 0,
                clientRect.right,
                topMargin + shift
            )
        );
    }
}

void CTEditView::ScrollDown(int n) {
    if (n == 0)
        return;
    windowY += n;

    if (n > windowWidth / 2) {
        Invalidate();
    } else {
        int shift = dy * n;
        CRect clientRect;
        GetClientRect(clientRect);
        CRect textRect( // Clippng rectangle
            leftMargin, topMargin,
            leftMargin + windowWidth*dx,
            topMargin + windowHeight*dy
        );
        ScrollWindow(
            0, -shift,
            textRect,   // Rectangle to scroll
            textRect    // Clippng rectangle
        );
        InvalidateRect(
            CRect(
                0,
                topMargin + windowHeight * dy - shift,
                clientRect.right,
                clientRect.bottom
            )
        );
    }
}

void CTEditView::onCharTyped() // Insert a character typed
{
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return;

    if (lastCharTyped == 0)
        return;

    if (cursorY == pDoc->text.size())
        onInsertLine();
    pDoc->text.setPointer(cursorY);
    TextLine& line = pDoc->text.getLine(cursorY);

    if (cursorX > line.length()) {
        int extraSpaces = cursorX - line.length();
        line.ensureCapacity(line.length() + extraSpaces + 2);
        while(--extraSpaces >= 0)
            line.append(' ');   // Add space to the end of line
    }
    line.insert(cursorX, lastCharTyped);
    line.trim();
    cursorX++;
    RedrawTextRectangle(cursorX - 1, cursorY, INT_MAX, 1);
}

void CTEditView::RedrawTextRectangle(
    int x, int y, int w, int h
) {
    int x1 = x + w;
    if (w == INT_MAX)
        x1 = INT_MAX;
    int y1 = y + h;
    if (h == INT_MAX)
        y1 = INT_MAX;

    int x0 = x;
    if (x0 < windowX)
        x0 = windowX;
    int y0 = y;
    if (y0 < windowY)
        y0 = windowY;

    if (x1 > windowX + windowWidth)
        x1 = windowX + windowWidth;
    if (y1 > windowY + windowHeight)
        y1 = windowY + windowHeight;
    if (x1 <= x || y1 <= y)
        return;

    int left = leftMargin + (x0 - windowX) * dx;
    int top = topMargin + (y0 - windowY) * dy;
    int width = (x1 - x0) * dx;
    int height = (y1 - y0) * dy;

    CRect clientRect;
    GetClientRect(clientRect);

    if (left + width > clientRect.Width())
        width -= (left + width - clientRect.Width());
    if (top + height > clientRect.Height())
        height -= (top + height - clientRect.Height());

    InvalidateRect(
        CRect(left, top, left + width, top + height)
    );
}

void CTEditView::onInsert()   // Insert a space
{
    lastCharTyped = ' ';
    onCharTyped();
    --cursorX;
}

void CTEditView::onInsertLine() // Insert an empty above the current
{
    CTEditDoc* pDoc = GetDocument();
    pDoc->text.setPointer(cursorY);
    pDoc->text.addAfter(new TextLine());
    RedrawTextRectangle(0, cursorY, INT_MAX, INT_MAX);
}

void CTEditView::onEnter()
{
    CTEditDoc* pDoc = GetDocument();
    pDoc->text.setPointer(cursorY);
    if (cursorY < pDoc->text.size()) {
        TextLine& line = pDoc->text.getLine(cursorY);
        pDoc->text.moveForward();
        int l = line.length();
        if (cursorX >= l) {
            pDoc->text.addBefore(new TextLine());
        } else {
            pDoc->text.addBefore(
                new TextLine(line.getString() + cursorX)
            );
            line.truncate(cursorX);
            line.trim();
        }
    } else {
        pDoc->text.addBefore(new TextLine());
    }
    cursorX = 0;
    ++cursorY;
    RedrawTextRectangle(0, cursorY - 1, INT_MAX, INT_MAX);
}

void CTEditView::onDelete() { // Delete current character
    CTEditDoc* pDoc = GetDocument();
    if (cursorY >= pDoc->text.size())
        return;
    TextLine& line = pDoc->text.getLine(cursorY);
    if (cursorX < line.length()) {
        line.removeAt(cursorX);
    }
    line.trim();
    RedrawTextRectangle(cursorX, cursorY, INT_MAX, 1);
}

void CTEditView::onDeleteLine() {
    CTEditDoc* pDoc = GetDocument();
    if (cursorY >= pDoc->text.size())
        return;
    pDoc->text.setPointer(cursorY);
    pDoc->text.removeAfter();
    RedrawTextRectangle(0, cursorY, INT_MAX, INT_MAX);
}

void CTEditView::onBackSpace() { // Delete previous character
    CTEditDoc* pDoc = GetDocument();
    if (cursorX <= 0 || cursorY >= pDoc->text.size())
        return;
    --cursorX;
    onDelete();
}

void CTEditView::onPageUp() {   // Page Up
    cursorY -= windowHeight;
    if (cursorY < 0)
        cursorY = 0;
}

void CTEditView::onPageDown() { // Page Down
    CTEditDoc* pDoc = GetDocument();
    cursorY += windowHeight;
    if (cursorY > pDoc->text.size())
        cursorY = pDoc->text.size();
}

void CTEditView::onQuit() { // Ask user, if text was changed, save, and quit
    if (processQuit())
    {
        //... PostQuitMessage(0);
        GetParent()->PostMessageW(WM_CLOSE);
    }
}

bool CTEditView::processQuit() { // Process the "Quit" request. Returns "true"
                                 // if window should be closed
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return false;

    if (pDoc->textChanged)
    {
        CSaveQueryDlg saveDlg;
        UINT res = saveDlg.DoModal();
        if (res == IDOK)
        {
            OnFileSave();
        }
        else if (res == IDCANCEL)
            return false;
    }
    pDoc->textChanged = false;  // To force quit
    return true;
}

void CTEditView::OnActivateFrame(UINT nState, CFrameWnd* pDeactivateFrame)
{
    // TODO: Add your specialized code here and/or call the base class
    inputDisabled = (nState == WA_INACTIVE);
    DrawCursor(cursorX, cursorY, !inputDisabled);

    CView::OnActivateFrame(nState, pDeactivateFrame);
}

void CTEditView::OnLButtonDown(UINT nFlags, CPoint point)
{
    // TODO: Add your message handler code here and/or call default
    CTEditDoc* pDoc = GetDocument();

    if (!inputDisabled && pDoc != NULL)
    {
        // Calculate the new position of cursor
        int x = point.x;
        int y = point.y;
        int cx = (x - leftMargin) / dx;
        if (cx >= windowWidth)
            cx = windowWidth - 1;
        cx += windowX;

        int cy = (y - topMargin) / dy;
        if (cy >= windowHeight)
            cy = windowHeight - 1;
        cy += windowY;
        if (cy > pDoc->text.size())
            cy = pDoc->text.size();

        // Erase the old curson and draw the new one
        DrawCursor(cursorX, cursorY, false);
        cursorX = cx; cursorY = cy;
        DrawCursor(cursorX, cursorY, true);
    }

    CView::OnLButtonDown(nFlags, point);
}

void CTEditView::onTab()
{
    ++cursorX;
    if ((cursorX % TAB_DX) != 0)
        cursorX += TAB_DX - (cursorX % TAB_DX);
}

void CTEditView::onTabLeft()
{
    if (cursorX > 0)
    {
        --cursorX;
        if ((cursorX % TAB_DX) != 0)
            cursorX -= (cursorX % TAB_DX);
    }
}

void CTEditView::onInsertTab() // Insert a tabulation (several spaces)
{
    CTEditDoc* pDoc = GetDocument();
    if (pDoc == NULL)
        return;

    if (cursorY == pDoc->text.size())
        onInsertLine();
    pDoc->text.setPointer(cursorY);
    TextLine& line = pDoc->text.getLine(cursorY);
    int lineLen = line.length();

    if (cursorX >= lineLen)
    {
        onTab();
    }
    else
    {
        char tab[16];
        int posX = cursorX;
        ++posX;
        int numSpaces = 1;
        if ((posX % TAB_DX) != 0)
        {
            int extraSpaces = TAB_DX - (posX % TAB_DX);
            posX += extraSpaces;
            numSpaces += extraSpaces;
        }
        if (numSpaces > 14)
            numSpaces = 14;
        for (int i = 0; i < numSpaces; ++i)
            tab[i] = ' ';
        tab[numSpaces] = 0;
        line.insert(cursorX, tab);
        RedrawTextRectangle(cursorX, cursorY, INT_MAX, 1);
        cursorX += numSpaces;
    }
}
